# frozen_string_literal: true

require "digest/sha2"
require "feen/parser"
require "ugoki/selector"

module Romance
  class Game
    def self.load(db, starting_position_feen)
      new(db, **::FEEN::Parser.call(starting_position_feen))
    end

    def initialize(db, **starting_position_params)
      @db = db

      @positions = [
        Position.new(**starting_position_params)
      ]
    end

    def play!(move)
      @positions << current_position.after(move) if legal?(*move)
    end

    def current_position
      @positions.fetch(-1)
    end

    def sha256
      ::Digest::SHA256.hexdigest(position_feens.join)
    end

    def over?
      !current_position.ally_king_on_the_board?
    end

    def inspect
      "<Romance::Game##{object_id} position ##{@positions.length}>"
    end

    private

    def position_feens
      @positions.map(&:to_feen).uniq.sort
    end

    def legal?(src_index, dst_index, moved_piece, captured = nil)
      raise Error::TheKingHasBeenCaptured unless current_position.ally_king_on_the_board?

      move = [src_index, dst_index, moved_piece, captured]

      raise Error::InvalidPseudoLegalMove unless pseudo_legal_moves(current_position).include?(move)

      next_position = current_position.after(move)

      raise Error::RepetitionOfPosition if repetition?(next_position)

      true
    end

    def repetition?(next_position)
      @positions.map(&:to_s).include?(next_position.to_s)
    end

    # Returns all pseudo-legal moves.
    #
    # @param position [Position] The state of the game.
    #
    # @return [Set] All pseudo-legal moves.
    def pseudo_legal_moves(position)
      ::Ugoki::Selector.new(@db, position).call
    end
  end
end

Dir[File.join(File.dirname(__FILE__), "error", "*.rb")].each do |fname|
  require_relative fname
end

require_relative "position"
